from typing import Optional, Tuple, Union

from ruxit.api.base_plugin import RemoteBasePlugin
import logging
import datetime
import json
import os
import os.path
from subprocess import Popen, PIPE
from ruxit.api.exceptions import ConfigException
import base64
import socket
import tempfile
import requests
from pymmh3 import hash128
import time

logger = logging.getLogger(__name__)

class SAPPluginRemote(RemoteBasePlugin):
    # Used to always have debug logging enabled the first run
    firstRun = True

    # Used to poll the data from where the previous poll ended
    lastPoll = None
    pollTo = None
    
    # Used to write temporary files
    tempdir = None

    # Used for not spamming events
    previousEvents = []
    raisedEvents = []
    
    # Used for limiting events
    MAX_EVENTS = 99
    events = 0
    
    # Plugin configuration
    username = None
    password = None
    server = None
    instance = None
    clientno = None
    workfolder = None
    version = None
    destinationFile = ""
    metrics = True
    usems = False
    mssysname = ""
    mshost = ""
    msport = ""
    msgroup = ""
    useTerminalForIp = False
    thresholds = False
    tasks = ""
    tasksToMonitor = []
    rum = False
    rfc = False
    rfcfunc = False
    apiToken = ""
    beaconURL = ""
    environmentID = ""
    applicationID = ""
    proxy = ""
    capUser = True
    timeoutMin = 25
    useHostName = False
    groups = {}
    devices = {}
    pollEveryXMin = 1
    pollCounter = 0
    payload = []
    
    def query(self, **kwargs):
        self.payload = []
        config = kwargs['config']
        if "pollLess" in config:
            if config["pollLess"]:
                self.pollEveryXMin = 5
            else:
                self.pollEveryXMin = 1
        self.groups = {}
        self.devices = {}
        self.username = config["username"].strip()
        self.password = config["password"].strip()
        self.server = config["server"].strip()
        self.instance = config["instance"].strip()
        self.clientno = config["clientno"].strip()
        if "metrics" in config:
            self.metrics = config["metrics"]
        if "usems" in config:
            self.usems = config["usems"]
        if "mssysname" in config:
            self.mssysname = config["mssysname"].strip()
        if "mshost" in config:
            self.mshost = config["mshost"].strip()
        if "msport" in config:
            self.msport = config["msport"].strip()
        if "msgroup" in config:
            self.msgroup = config["msgroup"].strip()
        if "threshold" in config:
            self.thresholds = config["threshold"]
        if "useTerminalForIp" in config:
            self.useTerminalForIp = config["useTerminalForIp"]
        pollAll = True
        if "pollAll" in config:
            pollAll = config["pollAll"]
        if "tasks" in config:
            self.tasks = config["tasks"]
        if "destinationFile" in config:
            self.destinationFile = config["destinationFile"].strip()
        self.workfolder = config["workfolder"].strip().replace("\\", "/")
        if "rum" in config:
            self.rum = config["rum"]
        if "rfc" in config:
            if config["rfc"] == "Do not report":
                self.rfc = False
                self.rfcfunc = False
            elif config["rfc"] == "Per remote destination":
                self.rfc = True
                self.rfcfunc = False
            elif config["rfc"] == "Per function call":
                self.rfc = False
                self.rfcfunc = True
        if "beaconURL" in config:
            self.beaconURL = config["beaconURL"].strip()
        if "environmentID" in config:
            self.environmentID = config["environmentID"].strip()
            self.mintURL = f"https://localhost/e/{self.environmentID}/api/v2/metric/ingest"
            if self.beaconURL == "":
                self.beaconURL = "https://localhost:9999/mbeacon/" + self.environmentID
        if "apiToken" in config:
            self.apiToken = config["apiToken"].strip()
        if (self.rfc or self.rfcfunc) and self.apiToken == "":
            raise ConfigException("Please enter an API token")
        if "applicationID" in config:
            self.applicationID = config["applicationID"].strip()
        if "proxy" in config:
            self.proxy = config["proxy"]
        if "capUser" in config:
            self.capUser = config["capUser"]
        if "timeoutMin" in config:
            self.timeoutMin = config["timeoutMin"]
        if self.timeoutMin < 5:
            raise ConfigException("Please enter a timeout between 5 and 25 minutes")
        if self.timeoutMin > 25:
            self.timeoutMin = 25
        if "useHostName" in config:
            self.useHostName = config["useHostName"]
        if not self.metrics and not self.rum and not self.rfc and not self.rfcfunc:
            raise ConfigException("None of user sessions, RFC or metrics have been selected")
        if (self.rum or self.rfc or self.rfcfunc) and (self.beaconURL == "" or self.beaconURL == "https://localhost:9999/mbeacon/"):
            raise ConfigException("No environment ID entered")
        if self.rum and self.applicationID == "":
            raise ConfigException("No application ID entered")
        if self.destinationFile == "" and not self.usems:
            if len(self.instance) != 2 or len([x for x in self.instance if x.isdigit()]) != 2:
                raise ConfigException("The instance id needs to be two digits")
        if self.destinationFile != "":
            if not os.path.isfile(self.destinationFile):
                raise ConfigException("Cannot find configured jCo destination file")
                
        javaPath = ""
        if "javaPath" in config:
            javaPath = config["javaPath"].strip()
        if javaPath == "":
            javaPath = self.getJavaPath()
            
            
        if pollAll:
            self.useHostName = False
        self.version = kwargs["json_config"]["version"]
        self.events = 0
        self.activeThreads = 0
        self.raisedEvents = []
        
        subFolder = self.server + self.instance + self.clientno
        if self.usems:
            subFolder = self.mshost + self.mssysname + self.msgroup + self.clientno
        # Make sure that there are no illegal characters in the folder name
        keepcharacters = (' ','.','_')
        subFolder = "".join(c for c in subFolder if c.isalnum() or c in keepcharacters).rstrip()
        
        self.tempdir = tempfile.gettempdir() + "/" + subFolder + "/"
        
        # Create the folder if it doesn't exist
        if not os.path.exists(self.tempdir):
            os.makedirs(self.tempdir)
            
        if not self.workfolder.endswith("/"):
            self.workfolder += "/"
        
        debugLogging = config["debug"]
        if debugLogging or self.firstRun:
            debugJava = True
            logger.setLevel(logging.DEBUG)
        else:
            debugJava = False
            logger.setLevel(logging.WARNING)
            
        self.tasksToMonitor = []
        for task in self.tasks.split(","):
            self.tasksToMonitor.append(task.strip())
            logger.debug(f"Monitoring {task.strip()}")
            
        poll = True
        if not self.firstRun:
            self.pollCounter += 1
            if self.pollEveryXMin > self.pollCounter:
                logger.setLevel(logging.DEBUG)
                logger.info(f"Skipping execution {self.pollCounter}, executing every {self.pollEveryXMin} minutes")
                poll = False
            else:
                self.pollCounter = 0
        else:
            self.pollCounter = 0
        if poll:
            # Pass parameters to the java application
            currentdate = datetime.datetime.now()
            if self.lastPoll != None:
                pollFrom = max(self.lastPoll, currentdate - datetime.timedelta(minutes=2 + self.pollEveryXMin * 3)) #Poll a max of 2 intervals
            else:
                pollFrom = currentdate - datetime.timedelta(minutes=3+self.pollEveryXMin)
            self.pollTo = currentdate - datetime.timedelta(minutes=3)
            fromDate = pollFrom.strftime('%Y%m%d')
            toDate = self.pollTo.strftime('%Y%m%d')
            fromTime = pollFrom.strftime('%H%M') + "00"
            toTime = self.pollTo.strftime('%H%M') + "00"
            parameters = {"server": self.server, 
                "instance": self.instance, 
                "clientno": self.clientno, 
                "username": self.username, 
                "password": str(hash(self.password)), 
                "fromDate": fromDate, 
                "fromTime": fromTime, 
                "toDate": toDate, 
                "toTime": toTime,
                "debugLogging": debugJava,
                "RUM": self.rum,
                "beacon": self.beaconURL,
                "destinationFile": self.destinationFile,
                "pollAll": pollAll,
                "applicationID": self.applicationID,
                "proxy": self.proxy,
                "capUser": self.capUser,
                "useTerminalForIp": self.useTerminalForIp,
                "timeoutMin": self.timeoutMin,
                "pluginVersion": self.version,
                "usems": self.usems,
                "mssysname": self.mssysname,
                "mshost": self.mshost,
                "msport": self.msport,
                "msgroup": self.msgroup,
                "pollEveryXMin": self.pollEveryXMin}
            logger.info(parameters)
            
            confFile = self.tempdir + "config.dt.sap"
            logger.info("Write data to " + confFile)
            try:
                json_file = open(confFile, "w")
                json_file.write(json.dumps(parameters))
                json_file.close()
            except Exception as e:
                logger.exception(e)
        startedJava = False
        try:
            if os.name != 'nt':
                import fcntl
                import struct
                f = open(self.tempdir + "javaRunning.dt.sap.lock", 'w')
                lockdata = struct.pack('hhllhh', fcntl.F_WRLCK, 0, 0, 0, 0, 0)
                unlockdata = struct.pack('hhllhh', fcntl.F_UNLCK, 0, 0, 0, 0, 0)
                fcntl.fcntl(f.fileno(), fcntl.F_SETLK, lockdata)
                fcntl.fcntl(f.fileno(), fcntl.F_SETLK, unlockdata)
            os.unlink(self.tempdir + "javaRunning.dt.sap.lock")
            if self.firstRun:
                logger.info("File not in use, start process")
            else:
                logger.warn("File not in use, start process")
            if self.runJava(javaPath):
                startedJava = True
        except IOError as e:
            if e.errno == 2:
                if self.firstRun:
                    logger.info("No file found, start process")
                else:
                    logger.warn("No file found, start process")
                if self.runJava(javaPath):
                    startedJava = True
            if e.errno != 11 and e.errno != 13: # Resource in use
                logger.info(e)
        
        if startedJava and not self.firstRun:
            self.firstRun = True
        else:
            self.firstRun = False
            
        # Look for new data
        foundData = False
        for filename in os.listdir(self.tempdir):
            if filename.endswith(".dt.sap.res"):
                try:
                    logger.info("Get data from " + self.tempdir + filename)
                    results = None
                    with open(self.tempdir + filename, "r") as read_file:
                        try:
                            results = json.load(read_file)
                        except Exception as e:
                            logger.exception(e)
                            lines = read_file.readlines()
                            for line in lines:
                                logger.warning(line)
                    os.unlink(self.tempdir + filename)
                    if results is not None:
                        if self.processData(results, True):
                            self.lastPoll = self.pollTo
                        foundData = True
                except Exception as e:
                    logger.exception(e)

        if not foundData:
            logger.info("No previous data found, awaiting next execution")
        self.previousEvents = []
        self.previousEvents.extend(self.raisedEvents)

    def getJavaPath(self):
        javaPath = "java"
        if os.path.isfile("../../../gateway/jre/bin/java.exe"):
            javaPath = "../../../gateway/jre/bin/java.exe"
        elif os.path.isfile("../../../gateway/jre/bin/java"):
            javaPath = "../../../gateway/jre/bin/java"
        elif os.path.isfile("../../../jre/bin/java.exe"):
            javaPath = "../../../jre/bin/java.exe"
        elif os.path.isfile("../../../jre/bin/java"):
            javaPath = "../../../jre/bin/java"
        elif 'GATEWAY_HOME' in os.environ:
            if os.path.isfile(os.environ['GATEWAY_HOME'] + "/jre/bin/java.exe"):
                javaPath = os.environ['GATEWAY_HOME'] + "/jre/bin/java.exe"
            elif os.path.isfile(os.environ['GATEWAY_HOME'] + "/jre/bin/java"):
                javaPath = os.environ['GATEWAY_HOME'] + "/jre/bin/java"
        return javaPath

    def runJava(self, javaPath):
        logger.info("Using java path: " + javaPath)
        separator = ":"
        if os.name == 'nt':
            separator = ";"
        
        my_env = os.environ.copy()
        my_env["lp"] = str(base64.b64encode(self.password.encode('utf-8')), "utf-8")
        startup = [javaPath, "-Xmx1536m", "-cp", os.path.dirname(__file__) + "/SAPRFC.jar" + separator + self.workfolder + "sapjco3.jar", "com.dynatrace.saprfc.Main", self.tempdir]
        logger.info(startup)
        if not os.path.isfile(self.workfolder + "sapjco3.jar"):
            raise ConfigException("Missing 'sapjco3.jar' in '" + self.workfolder + "', please place the file in the work folder")
        if os.name == 'nt':
            if not os.path.isfile(self.workfolder + "sapjco3.dll"):
                raise ConfigException("Missing 'sapjco3.dll' in '" + self.workfolder + "', please place the file in the work folder")
            p = Popen(startup[:-1], env=my_env, shell=False, stdin=None, stdout=PIPE, stderr=PIPE)
            output, error = p.communicate(timeout=30)
            if b"No arguments found" not in output:
                logger.error("Starting java failed ret='%d' output='%s' error='%s'" % (p.returncode, output, error))
                return False
            DETACHED_PROCESS = 0x00000008
            Popen(startup, env=my_env, shell=False, stdin=None, stdout=None, stderr=None, close_fds=True, creationflags=DETACHED_PROCESS)
        else:
            if not os.path.isfile(self.workfolder + "libsapjco3.so"):
                raise ConfigException("Missing 'libsapjco3.so' in '" + self.workfolder + "', please place the file in the work folder")
            p = Popen(startup[:-1], shell=False, stdin=None, stdout=PIPE, stderr=PIPE)
            output, error = p.communicate(timeout=30)
            if b"No arguments found" not in output:
                logger.error("Starting java failed ret='%d' output='%s' error='%s'" % (p.returncode, output, error))
                return False
            startup.insert(0, "nohup")
            Popen(startup, env=my_env, shell=False, stdin=None, stdout=None, stderr=None, close_fds=True, preexec_fn=os.setpgrp )
        
    def processData(self, result, originalData):
        try:
            hostname = socket.gethostbyaddr(self.server)[0]
        except Exception as e:
            hostname = self.server
        if originalData:
            if "exception" in result:
                logger.warning("RFC exception: " + result["exception"])
            else:
                previousLevel = logger.level
                logger.setLevel(logging.DEBUG)
                for logline in result["logging"]:
                    if logline.strip() != "":
                        logger.info("RFC logging: " + logline)
                logger.setLevel(previousLevel)
        
        if self.rfc or self.rfcfunc or self.metrics:
            if "exception" in result:
                return False

        for instances in result["Results"]:
            for instance, details in instances.items():
                try:
                    logger.info(f"Polling {details['SID']} - {details['Host']} ({details['SNR']})")
                    serverName = f"{details['Host']} ({details['SNR']})"
                    host = details['Host']
                    if self.useHostName:
                        serverName = self.server
                        host = self.server
                    groupID = "SAP - " + details["SID"]
                    deviceID = f"SAP - {details['SID']} - {serverName}"
                    if self.rfc or self.rfcfunc:
                        dt_group_id, dt_custom_device_id = dt_group_and_device_ids(groupID, deviceID)
                        entityID = dt_custom_device_id
                        if "Workload" in details and "OUTPUT" in details["Workload"]:
                            if self.rfc:
                                if "RFCSRVRDEST" in details["Workload"]["OUTPUT"] and "item" in details["Workload"]["OUTPUT"]["RFCSRVRDEST"]:
                                    rfcSrvrDest = {}
                                    for series in details["Workload"]["OUTPUT"]["RFCSRVRDEST"]["item"]:
                                        rfcItem = series
                                        if isinstance(rfcItem, str):
                                            rfcItem = details["Workload"]["OUTPUT"]["RFCSRVRDEST"]["item"]
                                        if rfcItem["REMOT_DEST"] not in rfcSrvrDest:
                                            rfcSrvrDest[rfcItem["REMOT_DEST"]] = {"EXE_TIME":0, "CALL_TIME":0, "RECEIVE": 0, "SEND": 0, "COUNTER": 0}
                                        rfcSrvrDest[rfcItem["REMOT_DEST"]]["EXE_TIME"] += float(rfcItem["EXE_TIME"])
                                        rfcSrvrDest[rfcItem["REMOT_DEST"]]["CALL_TIME"] += float(rfcItem["CALL_TIME"])
                                        rfcSrvrDest[rfcItem["REMOT_DEST"]]["RECEIVE"] += float(rfcItem["RECEIVE"])
                                        rfcSrvrDest[rfcItem["REMOT_DEST"]]["SEND"] += float(rfcItem["SEND"])
                                        rfcSrvrDest[rfcItem["REMOT_DEST"]]["COUNTER"] += float(rfcItem["COUNTER"])
                                        
                                    for rfcDest, rfcItem in rfcSrvrDest.items():
                                        dimensions = {"remote_destination": rfcDest, "dt.entity.custom_device": entityID, "sid": details["SID"], "instance_number": details['SNR'], "host": host}
                                        self.addMintMetric("rfcIncomingExecutionTime", dimensions, self.getGauge(rfcItem["EXE_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingCallTime", dimensions, self.getGauge(rfcItem["CALL_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingReceive", dimensions, self.getGauge(rfcItem["RECEIVE"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingSend", dimensions, self.getGauge(rfcItem["SEND"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingRequests", dimensions, str(rfcItem["COUNTER"]))
                                    
                                if "RFCCLNTDEST" in details["Workload"]["OUTPUT"] and "item" in details["Workload"]["OUTPUT"]["RFCCLNTDEST"]:
                                    rfcClntDest = {}
                                    for series in details["Workload"]["OUTPUT"]["RFCCLNTDEST"]["item"]:
                                        rfcItem = series
                                        if isinstance(rfcItem, str):
                                            rfcItem = details["Workload"]["OUTPUT"]["RFCCLNTDEST"]["item"]
                                        if rfcItem["REMOT_DEST"] not in rfcClntDest:
                                            rfcClntDest[rfcItem["REMOT_DEST"]] = {"EXE_TIME":0, "CALL_TIME":0, "RECEIVE": 0, "SEND": 0, "COUNTER": 0}
                                        rfcClntDest[rfcItem["REMOT_DEST"]]["EXE_TIME"] += float(rfcItem["EXE_TIME"])
                                        rfcClntDest[rfcItem["REMOT_DEST"]]["CALL_TIME"] += float(rfcItem["CALL_TIME"])
                                        rfcClntDest[rfcItem["REMOT_DEST"]]["RECEIVE"] += float(rfcItem["RECEIVE"])
                                        rfcClntDest[rfcItem["REMOT_DEST"]]["SEND"] += float(rfcItem["SEND"])
                                        rfcClntDest[rfcItem["REMOT_DEST"]]["COUNTER"] += float(rfcItem["COUNTER"])
                                        
                                    for rfcDest, rfcItem in rfcClntDest.items():
                                        dimensions = {"remote_destination": rfcDest, "dt.entity.custom_device": entityID, "sid": details["SID"], "instance_number": details['SNR'], "host": host}
                                        self.addMintMetric("rfcOutgoingExecutionTime", dimensions, self.getGauge(rfcItem["EXE_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingCallTime", dimensions, self.getGauge(rfcItem["CALL_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingReceive", dimensions, self.getGauge(rfcItem["RECEIVE"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingSend", dimensions, self.getGauge(rfcItem["SEND"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingRequests", dimensions, str(rfcItem["COUNTER"]))
                                    
                            if self.rfcfunc:
                                if "RFCSRVR" in details["Workload"]["OUTPUT"] and "item" in details["Workload"]["OUTPUT"]["RFCSRVR"]:
                                    for series in details["Workload"]["OUTPUT"]["RFCSRVR"]["item"]:
                                        rfcItem = series
                                        if isinstance(rfcItem, str):
                                            rfcItem = details["Workload"]["OUTPUT"]["RFCSRVR"]["item"]
                                        dimensions = {"remote_destination": rfcItem["REMOT_DEST"], "dt.entity.custom_device": entityID, "sid": details["SID"], "instance_number": details['SNR'], "host": host,
                                            "target": rfcItem["TARGET"], "program_name": rfcItem["PROG_NAME"], "local_destination": rfcItem["LOCAL_DEST"], "client_number": rfcItem["MANDT"], 
                                            "account": rfcItem["ACCOUNT"] if self.capUser else "", "user_id": rfcItem["USERID"] if self.capUser else "", "remote_destination": rfcItem["REMOT_DEST"], "function_name": rfcItem["FUNC_NAME"], "entry_id": rfcItem["ENTRY_ID"], 
                                            "rfc_caller": rfcItem["RFC_CALLER"], "task_type": self.TaskIdToTaskType(rfcItem["TASKTYPE"]).replace("一", "")
                                        }
                                        self.addMintMetric("rfcIncomingExecutionTime", dimensions, self.getGauge(rfcItem["EXE_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingCallTime", dimensions, self.getGauge(rfcItem["CALL_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingReceive", dimensions, self.getGauge(rfcItem["RECEIVE"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingSend", dimensions, self.getGauge(rfcItem["SEND"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcIncomingRequests", dimensions, str(rfcItem["COUNTER"]))
                                        
                                if "RFCCLNT" in details["Workload"]["OUTPUT"] and "item" in details["Workload"]["OUTPUT"]["RFCCLNT"]:
                                    for series in details["Workload"]["OUTPUT"]["RFCCLNT"]["item"]:
                                        rfcItem = series
                                        if isinstance(rfcItem, str):
                                            rfcItem = details["Workload"]["OUTPUT"]["RFCCLNT"]["item"]
                                        dimensions = {"remote_destination": rfcItem["REMOT_DEST"], "dt.entity.custom_device": entityID, "sid": details["SID"], "instance_number": details['SNR'], "host": host,
                                            "target": rfcItem["TARGET"], "program_name": rfcItem["PROG_NAME"], "local_destination": rfcItem["LOCAL_DEST"], "client_number": rfcItem["MANDT"], 
                                            "account": rfcItem["ACCOUNT"] if self.capUser else "", "user_id": rfcItem["USERID"] if self.capUser else "", "remote_destination": rfcItem["REMOT_DEST"], "function_name": rfcItem["FUNC_NAME"], "entry_id": rfcItem["ENTRY_ID"], 
                                            "task_type": self.TaskIdToTaskType(rfcItem["TASKTYPE"]).replace("一", "")
                                        }
                                        self.addMintMetric("rfcOutgoingExecutionTime", dimensions, self.getGauge(rfcItem["EXE_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingCallTime", dimensions, self.getGauge(rfcItem["CALL_TIME"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingReceive", dimensions, self.getGauge(rfcItem["RECEIVE"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingSend", dimensions, self.getGauge(rfcItem["SEND"], rfcItem["COUNTER"]))
                                        self.addMintMetric("rfcOutgoingRequests", dimensions, str(rfcItem["COUNTER"]))
                                    
                            self.sendMINTMetrics()
                    if self.metrics:
                        if details["SID"] in self.groups:
                            g1 = self.groups[details["SID"]]
                        else:
                            g1 = self.topology_builder.create_group(groupID, "SAP - ABAP - " + details["SID"])
                            self.groups[details["SID"]] = g1
                        if deviceID in self.devices:
                            e1 = self.devices[deviceID]
                        else:
                            e1 = g1.create_element(deviceID, deviceID)
                            self.devices[deviceID] = e1
                            e1.report_property("Plugin version", self.version)
                            
                            # If calling a single host using its DNS, convert it to an ip
                            hostIP = None
                            try:
                                socket.inet_aton(details["IP"]) # is the host name an ip
                                hostIP = details["IP"]
                            except socket.error as e:
                                try:
                                    hostIP = socket.gethostbyname(details["IP"])
                                except socket.error as e:
                                    logger.exception(e)
                                    
                            if hostIP is not None:
                                if not details["SNR"].isnumeric():
                                    if self.usems:
                                        e1.add_endpoint(hostIP, int(self.msport))
                                else:
                                    e1.add_endpoint(hostIP, int("32"+details["SNR"]))
                                    e1.add_endpoint(hostIP, int("33"+details["SNR"]))
                                    e1.add_endpoint(hostIP, int("3"+details["SNR"] + "15"))
                                    e1.add_endpoint(hostIP, int("5"+details["SNR"] + "13"))
                                    e1.add_endpoint(hostIP, int("5"+details["SNR"] + "14"))

                        e1.absolute(key='avail', value = details["Avb"])
                        
                        if details["Avb"] == 100:    
                            runtimeErrors = {
                                "Unclassified Error": 0,
                                "ABAP Programming Error": 0,
                                "ITS Error": 0,
                                "Error in Database Interface": 0,
                                "External Error": 0,
                                "Installation Errors": 0,
                                "Internal Kernel Error": 0,
                                "Internal ST Error": 0,
                                "No Error": 0,
                                "Resource Shortage": 0,
                                "Error in Open SQL": 0,
                                "Error at ABAP Runtime": 0,
                                "ST Runtime Error": 0,
                                "End User Error": 0,
                                "XSLT Runtime Error": 0,
                                "Error at Screen Runtime": 0
                            }
                            if "Events" in details:
                                for event in details["Events"]:
                                    if originalData:
                                        logger.info("Found event: " + str(event))
                                        self.report_custom_info_event(e1, "Runtime error: " + event["FIELD1"], event["FIELD1"], event)
                                    runtimeErrors[event["Category"]] +=1
                            for category in runtimeErrors:
                                e1.absolute(key='runtimeErrorsByCat', value = runtimeErrors[category], dimensions = {"Category": category})
                                
                            all_other_metrics = {}
                            all_other_metrics["count"] = 0
                            all_other_metrics["respTime"] = 0
                            all_other_metrics["avgcputime"] = 0
                            all_other_metrics["avgdbtime"] = 0
                            all_other_metrics["dbcalls"] = 0
                            all_other_metrics["rollintime"] = 0 
                            all_other_metrics["rollwaittime"] = 0
                            all_other_metrics["loadgentime"] = 0
                            totalRequests = 0
                            totalRespTime = 0
                            totalCputime = 0
                            totalDbtime = 0
                            totalRollwaittime = 0
                            totalLoadgentime = 0
                            if "Sessions" in details:
                                e1.absolute(key='sessions', value=int(details["Sessions"]))
                            if "Workload" in details and "OUTPUT" in details["Workload"] and "TASKTIMES" in details["Workload"]["OUTPUT"] and "item" in details["Workload"]["OUTPUT"]["TASKTIMES"]:
                                for series in details["Workload"]["OUTPUT"]["TASKTIMES"]["item"]:
                                    taskItem = series
                                    if isinstance(taskItem, str):
                                        taskItem = details["Workload"]["OUTPUT"]["TASKTIMES"]["item"]
                                    taskname = self.TaskIdToTaskType(taskItem["TASKTYPE"])
                                    dimensions = {'tasktype': taskname}
                                    respTime = float(taskItem["RESPTI"])/float(taskItem["COUNT"])
                                    avgcputime = float(taskItem["CPUTI"])/float(taskItem["COUNT"])
                                    avgdbtime = (float(taskItem["READDIRTI"]) + float(taskItem["READSEQTI"]) + float(taskItem["CHNGTI"]))/float(taskItem["COUNT"])
                                    dbcalls = taskItem["READSEQCNT"] + taskItem["CHNGCNT"] + taskItem["PHYREADCNT"] + taskItem["PHYCALLS"]
                                    totalRequests += float(taskItem["COUNT"])
                                    totalRespTime += float(taskItem["RESPTI"])
                                    totalCputime += float(taskItem["CPUTI"])
                                    totalDbtime += float(taskItem["READDIRTI"]) + float(taskItem["READSEQTI"]) + float(taskItem["CHNGTI"])
                                    
                                    if taskname == 'DIALOG':
                                        e1.absolute(key='guiresponsetime', value=float(respTime), dimensions=dimensions)
                                        e1.absolute(key='guidialogsteps', value=float(taskItem["COUNT"])/60, dimensions=dimensions)
                                    if taskname == '一OTHER TASKS' or taskname not in self.tasksToMonitor:
                                        all_other_metrics["count"] += taskItem["COUNT"]
                                        all_other_metrics["respTime"] += float(taskItem["RESPTI"])
                                        all_other_metrics["avgcputime"] += float(taskItem["CPUTI"])
                                        all_other_metrics["avgdbtime"] += (float(taskItem["READDIRTI"]) + float(taskItem["READSEQTI"]) + float(taskItem["CHNGTI"]))
                                        all_other_metrics["dbcalls"] += dbcalls
                                    else:
                                        e1.absolute(key='Task.dialogsteps', value=float(taskItem["COUNT"])/60, dimensions=dimensions)
                                        e1.absolute(key='Task.avgtottime', value=float(respTime), dimensions=dimensions)
                                        e1.absolute(key='Task.cputime', value=float(avgcputime), dimensions=dimensions)
                                        e1.absolute(key='Task.dbtime', value=float(avgdbtime), dimensions=dimensions)
                                        e1.absolute(key='Task.dbcalls', value=int(dbcalls), dimensions=dimensions)

                            if "Workload" in details and "OUTPUT" in details["Workload"] and "TASKTYPE" in details["Workload"]["OUTPUT"] and "item" in details["Workload"]["OUTPUT"]["TASKTYPE"]:
                                for series in details["Workload"]["OUTPUT"]["TASKTYPE"]["item"]:
                                    taskItem = series
                                    if isinstance(taskItem, str):
                                        taskItem = details["Workload"]["OUTPUT"]["TASKTYPE"]["item"]
                                    taskname = self.TaskIdToTaskType(taskItem["TASKTYPE"])
                                    dimensions = { 'tasktype': taskname }
                                    avgrollin = float(taskItem["ROLLINTI"])/float(taskItem["COUNT"])
                                    avgrollwait = float(taskItem["ROLLWAITTI"])/float(taskItem["COUNT"])
                                    avgloadgentime = float(taskItem["LOADGENTI"])/float(taskItem["COUNT"])
                                    avgguinettime = float(taskItem["GUINETTIME"])/float(taskItem["COUNT"])
                                    totalRollwaittime += float(taskItem["ROLLWAITTI"])
                                    totalLoadgentime += float(taskItem["LOADGENTI"])
                                        
                                    if taskname == 'DIALOG':
                                        e1.absolute(key='guinetworktime', value=avgguinettime, dimensions=dimensions)
                                    if taskname == '一OTHER TASKS' or taskname not in self.tasksToMonitor:
                                        all_other_metrics["rollintime"] += float(taskItem["ROLLINTI"])
                                        all_other_metrics["rollwaittime"] += float(taskItem["ROLLWAITTI"])
                                        all_other_metrics["loadgentime"] += float(taskItem["LOADGENTI"])
                                    else:
                                        e1.absolute(key='Task.rollintime', value=float(avgrollin), dimensions=dimensions)
                                        e1.absolute(key='Task.rollwaittime', value=float(avgrollwait), dimensions=dimensions)
                                        e1.absolute(key='Task.loadgentime', value=avgloadgentime, dimensions=dimensions)
                                
                            users = {}
                            usage = {}
                            if "DIALOG" in self.tasksToMonitor:
                                users["DIALOG"] = []
                                usage["DIALOG"] = 0
                            if "RFC" in self.tasksToMonitor:
                                users["RFC"] = []
                                usage["RFC"] = 0
                            if "Workload" in details and "OUTPUT" in details["Workload"] and "USERWORKLOAD" in details["Workload"]["OUTPUT"] and "item" in details["Workload"]["OUTPUT"]["USERWORKLOAD"]:
                                for series in details["Workload"]["OUTPUT"]["USERWORKLOAD"]["item"]:
                                    userItem = series
                                    if isinstance(userItem, str):
                                        userItem = details["Workload"]["OUTPUT"]["USERWORKLOAD"]["item"]
                                    taskname = self.TaskIdToTaskType(userItem["TASKTYPE"])
                                    if taskname in users:
                                        if userItem["USERNAME"] not in users[taskname]:
                                            users[taskname].append(userItem["USERNAME"])
                                            usage[taskname] += 1
                                    
                            for taskname in usage:
                                dimensions = { 'tasktype': taskname }
                                if taskname == "DIALOG":
                                    e1.absolute(key='guiusercount', value=int(usage[taskname]), dimensions=dimensions)
                                e1.absolute(key='Task.usercount', value=int(usage[taskname]), dimensions=dimensions)

                            if all_other_metrics["count"] > 0:
                                dimensions = { 'tasktype': '一OTHER TASKS' }
                                avgtottime_ao = float(all_other_metrics["respTime"])/float(all_other_metrics["count"])
                                avgcputime_ao = float(all_other_metrics["avgcputime"])/float(all_other_metrics["count"])
                                avgdbtime_ao = float(all_other_metrics["avgdbtime"])/float(all_other_metrics["count"])
                                loadgentime = float(all_other_metrics["loadgentime"])/float(all_other_metrics["count"])
                                dbcalls = float(all_other_metrics["dbcalls"])/len(all_other_metrics.keys())
                                e1.absolute(key='Task.dialogsteps', value=float(all_other_metrics["count"])/60, dimensions=dimensions)
                                e1.absolute(key='Task.avgtottime', value=avgtottime_ao, dimensions=dimensions)
                                e1.absolute(key='Task.cputime', value=avgcputime_ao, dimensions=dimensions)
                                e1.absolute(key='Task.dbtime', value=avgdbtime_ao, dimensions=dimensions)
                                e1.absolute(key='Task.dbcalls', value=dbcalls, dimensions=dimensions)
                                e1.absolute(key='Task.loadgentime', value=loadgentime, dimensions=dimensions)
                                e1.absolute(key='Task.rollintime', value=float(all_other_metrics["rollintime"])/float(all_other_metrics["count"]), dimensions=dimensions)
                                e1.absolute(key='Task.rollwaittime', value=float(all_other_metrics["rollwaittime"])/float(all_other_metrics["count"]), dimensions=dimensions)
                                
                            if totalRespTime > 0 and totalRequests > 0:
                                avgResponseTime = totalRespTime / totalRequests
                                cpuratio = int(100 * totalCputime / totalRespTime)
                                dbratio = int(100 * totalDbtime / totalRespTime)
                                waittimeratio = int(100 * totalRollwaittime / totalRespTime)
                                loadtimeratio = int(100 * totalLoadgentime / totalRespTime)
                                e1.absolute(key='cpuratio', value=cpuratio)
                                e1.absolute(key='dbtimeratio', value=dbratio)
                                e1.absolute(key='waittimeratio', value=waittimeratio)
                                e1.absolute(key='loadtimeratio', value=loadtimeratio)
                                e1.absolute(key='avgtime', value=avgResponseTime)
                            if self.thresholds:
                                e1.absolute(key='fiftyline', value=50)
                                e1.absolute(key='tenline', value=10)
                                e1.absolute(key='twoline', value=2)
                except Exception as e:
                    logger.exception(e)
                    logger.warning(details)
        return True

    def TaskIdToTaskType(self, tasktype):
        taskid = base64.b64decode(tasktype)
        if taskid == b'\x00':
            return "NONE"
        if taskid == b'\x01':
            return "DIALOG"
        if taskid == b'\x02':
            return "UPDATE"
        if taskid == b'\x03':
            return "SPOOL"
        if taskid == b'\x04':
            return "BCKGRD"
        if taskid == b'\x05':
            return "ENQUEUE"
        if taskid == b'\x06':
            return "BUF.SYN"
        if taskid == b'\x07':
            return "AUTOABA"
        if taskid == b'\x08':
            return "UPDATE2"
        if taskid == b'\x09':
            return "NATIVE_RFC"
        if taskid == b'\x0A':
            return "PLUGIN"
        if taskid == b'\x0B':
            return "AUTOTH"
        if taskid == b'\x0C':
            return "RPCTH"
        if taskid == b'\x0D':
            return "RFCVMC"
        if taskid == b'\x0E':
            return "DDLOG CLEANUP"
        if taskid == b'\x0F':
            return "DEL. THCALL"
        if taskid == b'\x10':
            return "AUTOJAVA"
        if taskid == b'\x11':
            return "LICENCESRV"
        if taskid == b'\x12':
            return "AUTOCCMS"
        if taskid == b'\x13':
            return "MSADM"
        if taskid == b'\x14':
            return "SYS_STARTUP"
        if taskid == b'\x15':
            return "BGRFCSDL"
        if taskid == b'\x16':
            return "BGRFC"
        if taskid == b'\x17':
            return "APC"
        if taskid == b'\x21':
            return "OTHERS"
        if taskid == b'\x22':
            return "DINOGUI"
        if taskid == b'\x23':
            return "B.INPUT"
        if taskid == b'\x65':
            return "HTTP"
        if taskid == b'\x66':
            return "HTTPS"
        if taskid == b'\x67':
            return "NNTP"
        if taskid == b'\x68':
            return "SMTP"
        if taskid == b'\x69':
            return "FTP"
        if taskid == b'\x6C':
            return "LCOM"
        if taskid == b'\x75':
            return "HTTP/JSP"
        if taskid == b'\x76':
            return "HTTPS/JSP"
        if taskid == b'\xF9':
            return "AUTO_RFC"
        if taskid == b'\xFA':
            return "WS-RFC"
        if taskid == b'\xFB':
            return "WS-HTTP"
        if taskid == b'\xFC':
            return "ESI"
        if taskid == b'\xFD':
            return "ALE"
        if taskid == b'\xFE':
            return "RFC"
        if taskid == b'\xFF':
            return "CPIC"
        return "一OTHER TASKS"
        
    def getGauge(self, val, count):
        return "gauge,min=" + str(val / count) + ",max=" + str(val / count) + ",sum=" + str(val) + ",count=" + str(round(count))
        
    def addMintMetric(self, metricName, dimensions, value):
        dimStr = ""
        # The SAP server is always polled in the past to give it time to finish processing
        timestamp = int(time.time() * 1000 - 4 * 60 * 1000)
        for dimKey, dimValue in dimensions.items():
            if dimValue != "":
                if isinstance(dimValue, str):
                    sanitizedValue = dimValue.replace("\\n", "").replace("\\t", "").replace("\\r", "").replace("\\", "").strip()
                    dimStr += f",{dimKey}=\"{sanitizedValue}\""
                else:
                    dimStr += f",{dimKey}={dimValue}"
        self.payload.append(f"sap.{metricName}{dimStr} {value} {timestamp}")
            
    def sendMINTMetrics(self):
        data = []
        for line in self.payload:
            if len(data) == 1000:
                self.doPostRequest(data)
                data = []
            data.append(line)
        if len(data) > 0:
            self.doPostRequest(data)
                
    def doPostRequest(self, data):
        headers = {"Authorization": "Api-Token " + self.apiToken, "Content-Type": "text/plain"}
        response = requests.post(f"https://localhost:9999/e/{self.environmentID}/api/v2/metrics/ingest", verify=False, headers=headers, data='\n'.join(data))
        if response.status_code != 202:
            logger.warn("Metric Ingest API call failed. Input: " + '\n'.join(data) + " Output: " + response.text)
        else:
            logger.info("Metric Ingest API call successful Output: " + response.text)
        
    def report_custom_info_event(self, e1, message, title, properties):
        if self.events < self.MAX_EVENTS:
            if properties != None:
                if title + message + str(properties) not in self.previousEvents:
                    self.events += 1
                    e1.report_custom_info_event(message, title, properties)
                self.raisedEvents.append(title + message + str(properties))
            else:
                if title + message not in self.previousEvents:
                    self.events += 1
                    e1.report_custom_info_event(message, title)
                self.raisedEvents.append(title + message)


def dt_group_and_device_ids(group_name: str, device_name: str, namespace: Optional[str] = "") -> Tuple[str, str]:
    """
    Calculates the Group ID and Custom Device ID
    """
    group_id = dt_murmur3(build_dt_group_id(group_name, namespace))
    custom_device_id = dt_murmur3(build_dt_custom_device_id(group_id, device_name))

    return f"CUSTOM_DEVICE_GROUP-{group_id}", f"CUSTOM_DEVICE-{custom_device_id}"


def dt_murmur3(data: bytes) -> str:
    return f"{int(hash128(data, seed=0)):X}"[-16:]


def build_dt_custom_device_id(group_id: Union[str, bytes], custom_device_name: str):

    group_id = int(group_id, 16)

    final_message = []

    # Convert the length of the group_id to Big Endian and add the last 4 bytes
    group_id_len_big_endian = group_id.to_bytes(8, byteorder="big", signed=False)
    final_message.extend(group_id_len_big_endian[-8:])

    # Add the custom_device_id
    final_message.extend(bytes(custom_device_name, encoding="utf-8"))

    # Convert the length of the device_id to Big Endian and add the last 4 bytes
    device_id_len_big_endian = len(custom_device_name).to_bytes(8, byteorder="big", signed=False)
    final_message.extend(device_id_len_big_endian[-4:])

    return bytes(final_message)


def build_dt_group_id(group_name: Union[str, bytes], namespace: str = ""):
    final_message = []

    # Add the namespace
    final_message.extend(bytes(namespace, encoding="utf-8"))

    # Convert the length of the namespace to Big Endian and add the last 4 bytes
    namespace_len_big_endian = len(namespace).to_bytes(8, byteorder="big", signed=False)
    final_message.extend(namespace_len_big_endian[-4:])

    # Add the group id
    final_message.extend(bytes(group_name, encoding="utf-8"))

    # Convert the length of the group_id to Big Endian and add the last 4 bytes
    group_id_len_big_endian = len(group_name).to_bytes(8, byteorder="big", signed=False)
    final_message.extend(group_id_len_big_endian[-4:])

    return bytes(final_message)